#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

template<bool UseVectorization>
__global__ void adaptive_swish_scaling_kernel(
    const float* __restrict__ input,
    float* output,
    float scaling_factor,
    int rows,
    int cols) {
    
    if constexpr(UseVectorization) {
        // Vectorized version for large contiguous data
        const int tid = threadIdx.x;
        const int elements_per_thread = 4;
        const int block_elements = blockDim.x * elements_per_thread;
        int base_idx = blockIdx.x * block_elements + tid * elements_per_thread;
        const int N = rows * cols;
        
        extern __shared__ float shared_data[];
        
        if (base_idx < N) {
            float4* in_vec = (float4*)(&input[base_idx]);
            float4* shared_vec = (float4*)(&shared_data[tid * elements_per_thread]);
            
            if (base_idx + 3 < N) {
                *shared_vec = *in_vec;
            } else {
                for (int i = 0; i < elements_per_thread && (base_idx + i) < N; i++) {
                    shared_data[tid * elements_per_thread + i] = input[base_idx + i];
                }
            }
        }
        __syncthreads();
        
        if (base_idx < N) {
            float4 result;
            float* result_f = (float*)&result;
            
            #pragma unroll
            for (int i = 0; i < elements_per_thread && (base_idx + i) < N; i++) {
                float x = shared_data[tid * elements_per_thread + i];
                float sigmoid = 1.0f / (1.0f + expf(-x));
                result_f[i] = x * sigmoid * scaling_factor;
            }
            
            if (base_idx + 3 < N) {
                float4* out_vec = (float4*)(&output[base_idx]);
                *out_vec = result;
            } else {
                for (int i = 0; i < elements_per_thread && (base_idx + i) < N; i++) {
                    output[base_idx + i] = result_f[i];
                }
            }
        }
    } else {
        // 2D version for better cache locality with matrix operations
        int row = blockIdx.y * blockDim.y + threadIdx.y;
        int col = blockIdx.x * blockDim.x + threadIdx.x;
        
        if (row < rows && col < cols) {
            int idx = row * cols + col;
            float x = input[idx];
            float sigmoid = 1.0f / (1.0f + expf(-x));
            output[idx] = x * sigmoid * scaling_factor;
        }
    }
}

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor weight,
    torch::Tensor bias,
    double scaling_factor) {
    
    x = x.contiguous();
    weight = weight.contiguous();
    bias = bias.contiguous();
    
    TORCH_CHECK(x.is_cuda(), "Input tensor 'x' must be a CUDA tensor.");
    TORCH_CHECK(weight.is_cuda(), "Weight tensor must be a CUDA tensor.");
    TORCH_CHECK(bias.is_cuda(), "Bias tensor must be a CUDA tensor.");
    TORCH_CHECK(x.scalar_type() == at::kFloat, "Input tensor 'x' must be of type torch.float32.");
    
    auto y = at::addmm(bias, x, weight.t());
    auto output = at::empty_like(y);
    
    const int rows = y.size(0);
    const int cols = y.size(1);
    const int total_elements = rows * cols;
    
    // Choose kernel configuration based on input size and shape
    if (cols >= 512 && rows == 1) {
        // Use vectorized version for large 1D operations
        const int threads = 256;
        const int elements_per_thread = 4;
        const int elements_per_block = threads * elements_per_thread;
        const int blocks = (total_elements + elements_per_block - 1) / elements_per_block;
        const size_t shared_mem_size = threads * elements_per_thread * sizeof(float);
        
        adaptive_swish_scaling_kernel<true><<<blocks, threads, shared_mem_size>>>(
            y.data_ptr<float>(),
            output.data_ptr<float>(),
            static_cast<float>(scaling_factor),
            rows,
            cols);
    } else {
        // Use 2D version for matrix operations
        dim3 threads(32, 32);
        dim3 blocks((cols + threads.x - 1) / threads.x, (rows + threads.y - 1) / threads.y);
        
        adaptive_swish_scaling_kernel<false><<<blocks, threads, 0>>>(
            y.data_ptr<float>(),
            output.data_ptr<float>(),
            static_cast<float>(scaling_factor),
            rows,
            cols);
    }
    
    cudaError_t err = cudaGetLastError();
    TORCH_CHECK(err == cudaSuccess, "CUDA kernel failed : ", cudaGetErrorString(err));
    
    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Adaptive CUDA forward function");
}